@@ -54,6 +54,12 @@ void uwsgi_destroy_processes() {
 
         uwsgi_detach_daemons();
 
+	for (i = 1; i <= uwsgi.numproc; i++) {
+		if (uwsgi.workers[i].pid > 0) {
+			waitpid(uwsgi.workers[i].pid, &waitpid_status, 0);
+		}
+	}
+
         for (i = 0; i < ushared->gateways_cnt; i++) {
                 if (ushared->gateways[i].pid > 0) {
                         kill(ushared->gateways[i].pid, SIGKILL);
@@ -1561,7 +1567,10 @@ void uwsgi_register_cheaper_algo(char *name, int (*func) (int)) {
 
 void trigger_harakiri(int i) {
 	int j;
-	uwsgi_log_verbose("*** HARAKIRI ON WORKER %d (pid: %d, try: %d) ***\n", i, uwsgi.workers[i].pid, uwsgi.workers[i].pending_harakiri + 1);
+	uwsgi_log_verbose("*** HARAKIRI ON WORKER %d (pid: %d, try: %d, graceful: %s) ***\n", i,
+				uwsgi.workers[i].pid,
+				uwsgi.workers[i].pending_harakiri + 1,
+				uwsgi.workers[i].pending_harakiri > 0 ? "no": "yes");
 	if (uwsgi.harakiri_verbose) {
 #ifdef __linux__
 		int proc_file;
@@ -1610,7 +1619,11 @@ void trigger_harakiri(int i) {
 		}
 
 		uwsgi_dump_worker(i, "HARAKIRI");
-		kill(uwsgi.workers[i].pid, SIGKILL);
+		if (uwsgi.workers[i].pending_harakiri == 0 && uwsgi.harakiri_graceful_timeout > 0) {
+			kill(uwsgi.workers[i].pid, uwsgi.harakiri_graceful_signal);
+		} else {
+			kill(uwsgi.workers[i].pid, SIGKILL);
+		}
 		if (!uwsgi.workers[i].pending_harakiri)
 			uwsgi.workers[i].harakiri_count++;
 		uwsgi.workers[i].pending_harakiri++;
